package il.ac.tau.cs.sw1.ex4;

import java.util.Arrays;
import java.util.Random;
import java.util.Scanner;

public class WordPuzzle {
	public static final char HIDDEN_CHAR = '_';
	
	/*
	 * @pre: template is legal for word
	 */
	public static char[] createPuzzleFromTemplate(String word, boolean[] template) { // Q - 1


		char[] puzzle = word.toCharArray();
		for (int i = 0; i < template.length; i++) {
			puzzle[i] = template[i] ? HIDDEN_CHAR : puzzle[i];
		}
		return puzzle;
	}

	public static boolean checkLegalTemplate(String word, boolean[] template) { // Q - 2

		if (word.length() != template.length) {
			return false;
		}
		
		boolean hasRevealed = false;
		boolean hasHidden = false;
		
		for (int i = 0; i < template.length; i++) {
			if (template[i]) {
				hasHidden = true;
			}
			
			if (!template[i]) {
				hasRevealed = true;
			}
			
			if (hasHidden && hasRevealed) {
				break;
			}
		}
		
		if (!hasRevealed || !hasHidden) {
			return false;
		}
		
		for (int i = 0; i < template.length; i++) {
			for (int j = 0; j < template.length; j++) {
				if (word.charAt(i) == word.charAt(j) && template[i] != template[j]) {
					return false;
				}
			}
		}
		
		return true;
	}
	
	/*
	 * @pre: 0 < k < word.lenght(), word.length() <= 10
	 */
	public static boolean[][] getAllLegalTemplates(String word, int k){  // Q - 3

		boolean[][] legalTemplates = new boolean[(int) Math.pow(2, word.length())][word.length()];
		int index = 0;

		for (int i = 0; i < Math.pow(2, word.length()); i++) {
			
			String binary = String.format("%1$" + word.length() + "s", Integer.toBinaryString(i))
					.replace(' ', '0');
			
			int trueCount = 0;
			boolean[] template = new boolean[word.length()];
			for (int j = 0; j < template.length; j++) {
				if (binary.charAt(j) == '1') {
					trueCount++;
					template[j] = true;
				}	
			}
			
			if (checkLegalTemplate(word, template) && trueCount == k) {
				legalTemplates[index++] = template;
			}
		}
		
		return Arrays.copyOf(legalTemplates, index);
	}
	
	
	/*
	 * @pre: puzzle is a legal puzzle constructed from word, guess is in [a...z]
	 */
	public static int applyGuess(char guess, String word, char[] puzzle) { // Q - 4

		int revealed = 0;
		for (int i = 0; i < puzzle.length; i++) {
			if (puzzle[i] == HIDDEN_CHAR && word.charAt(i) == guess) {
				revealed++;
				puzzle[i] = guess;
			}
		}
		return revealed;
	}
	

	/*
	 * @pre: puzzle is a legal puzzle constructed from word
	 * @pre: puzzle contains at least one hidden character. 
	 * @pre: there are at least 2 letters that don't appear in word, and the user didn't guess
	 */
	public static char[] getHint(String word, char[] puzzle, boolean[] already_guessed) { // Q - 5

		Random r = new Random();
		
		char[] hint = new char[2];
		
			
		char wrong;
		do {
			wrong = (char) (97 + r.nextInt(26));
		} while (already_guessed[(int) wrong - 97] || word.indexOf(wrong) > -1);
		
		char correct;
		
		do {
			correct = (char) (97 + r.nextInt(26));
		} while (already_guessed[(int) correct - 97] || word.indexOf(correct) == -1 || puzzle[word.indexOf(correct)] == correct);

		if (wrong < correct) {
			hint[0] = wrong;
			hint[1] = correct;
		} else {
			hint[0] = correct;
			hint[1] = wrong;
		}
		
		return hint;
	}

	

	public static char[] mainTemplateSettings(String word, Scanner inputScanner) { // Q - 6

		printSettingsMessage();
		
		boolean[] template = null;
		
		do {
			
			printSelectTemplate();
			
			String userChoice = inputScanner.next();
			
			if ("1".equals(userChoice)) {
				
				printSelectNumberOfHiddenChars();
				
				int hiddenChars = Integer.parseInt(inputScanner.next());
				
				boolean[][] allLegal = getAllLegalTemplates(word, hiddenChars);
				
				if (allLegal.length > 0) {
					Random r = new Random();
					template = allLegal[r.nextInt(allLegal.length)];
				}

			} else if ("2".equals(userChoice)) {
				
				printEnterPuzzleTemplate();
				
				
				String[] chars = inputScanner.next().split(",");
				
				template = new boolean[chars.length];
				
				for (int i = 0; i < template.length; i++) {
					template[i] = chars[i].equals(String.valueOf(HIDDEN_CHAR));
				}
				
				if (!checkLegalTemplate(word, template)) {
					template = null;
				}
			}
			
			if (template == null) {
				printWrongTemplateParameters();	
			}
			
		} while (template == null);
		
		
		
		return createPuzzleFromTemplate(word, template);
	}
	
	public static void mainGame(String word, char[] puzzle, Scanner inputScanner){ // Q - 7
		
		int hidden = 0;
		for (int i = 0; i < puzzle.length; i++) {
			if (puzzle[i] == HIDDEN_CHAR) {
				hidden++;
			}
		}
		
		int guesses = hidden + 3;
		
		boolean[] alreadyGuessed = new boolean[26];

		
		printGameStageMessage();
		
		printPuzzle(puzzle);
		
		while (true) {
			
			printEnterYourGuessMessage();
			
			String userChoice = inputScanner.next();
			
			if ("H".equals(userChoice)) {
				char[] hint = getHint(word, puzzle, alreadyGuessed);
				printHint(hint);
			} else {
				
				char guessed = userChoice.charAt(0);
				alreadyGuessed[(int) guessed - 97] = true;
				
				int revealed = applyGuess(guessed, word, puzzle);
				
				hidden -= revealed;
				
				if (revealed > 0) {
					if (hidden == 0) {
						printWinMessage();
						break;
					} else {
						guesses--;
						printCorrectGuess(guesses);
						printPuzzle(puzzle);
					}
					
				} else {
					
					guesses--;
					printWrongGuess(guesses);
				}
				
				if (guesses == 0) {
					printGameOver();
					break;
				}
				
			}
			
		}
		

	}
				
				


/*************************************************************/
/********************* Don't change this ********************/
/*************************************************************/

	public static void main(String[] args) throws Exception { 
		if (args.length < 1){
			throw new Exception("You must specify one argument to this program");
		}
		String wordForPuzzle = args[0].toLowerCase();
		if (wordForPuzzle.length() > 10){
			throw new Exception("The word should not contain more than 10 characters");
		}
		Scanner inputScanner = new Scanner(System.in);
		char[] puzzle = mainTemplateSettings(wordForPuzzle, inputScanner);
		mainGame(wordForPuzzle, puzzle, inputScanner);
		inputScanner.close();
	}


	public static void printSettingsMessage() {
		System.out.println("--- Settings stage ---");
	}

	public static void printEnterWord() {
		System.out.println("Enter word:");
	}
	
	public static void printSelectNumberOfHiddenChars(){
		System.out.println("Enter number of hidden characters:");
	}
	public static void printSelectTemplate() {
		System.out.println("Choose a (1) random or (2) manual template:");
	}
	
	public static void printWrongTemplateParameters() {
		System.out.println("Cannot generate puzzle, try again.");
	}
	
	public static void printEnterPuzzleTemplate() {
		System.out.println("Enter your puzzle template:");
	}


	public static void printPuzzle(char[] puzzle) {
		System.out.println(puzzle);
	}


	public static void printGameStageMessage() {
		System.out.println("--- Game stage ---");
	}

	public static void printEnterYourGuessMessage() {
		System.out.println("Enter your guess:");
	}

	public static void printHint(char[] hist){
		System.out.println(String.format("Here's a hint for you: choose either %s or %s.", hist[0] ,hist[1]));

	}
	public static void printCorrectGuess(int attemptsNum) {
		System.out.println("Correct Guess, " + attemptsNum + " guesses left.");
	}

	public static void printWrongGuess(int attemptsNum) {
		System.out.println("Wrong Guess, " + attemptsNum + " guesses left.");
	}

	public static void printWinMessage() {
		System.out.println("Congratulations! You solved the puzzle!");
	}

	public static void printGameOver() {
		System.out.println("Game over!");
	}

}